/**********************************************************************
Copyright (c) 2008 Andy Jefferson and others. All rights reserved.
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Contributors:
    ...
 **********************************************************************/
package org.datanucleus.store.ldap;

import java.lang.reflect.Constructor;
import java.lang.reflect.Method;
import java.lang.reflect.Modifier;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.HashSet;
import java.util.LinkedHashMap;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import javax.naming.ContextNotEmptyException;
import javax.naming.InvalidNameException;
import javax.naming.NameNotFoundException;
import javax.naming.NamingEnumeration;
import javax.naming.NamingException;
import javax.naming.directory.Attribute;
import javax.naming.directory.Attributes;
import javax.naming.directory.BasicAttributes;
import javax.naming.directory.DirContext;
import javax.naming.directory.SearchControls;
import javax.naming.directory.SearchResult;
import javax.naming.ldap.LdapName;
import javax.naming.ldap.Rdn;

import org.datanucleus.ClassLoaderResolver;
import org.datanucleus.FetchPlan;
import org.datanucleus.Transaction;
import org.datanucleus.exceptions.NucleusDataStoreException;
import org.datanucleus.exceptions.NucleusObjectNotFoundException;
import org.datanucleus.metadata.AbstractClassMetaData;
import org.datanucleus.metadata.AbstractMemberMetaData;
import org.datanucleus.metadata.MetaDataManager;
import org.datanucleus.metadata.MetaDataUtils;
import org.datanucleus.metadata.Relation;
import org.datanucleus.query.compiler.QueryCompilation;
import org.datanucleus.store.ExecutionContext;
import org.datanucleus.store.FieldValues;
import org.datanucleus.store.ObjectProvider;
import org.datanucleus.store.StoreManager;
import org.datanucleus.store.Type;
import org.datanucleus.store.connection.ManagedConnection;
import org.datanucleus.store.ldap.fieldmanager.AbstractMappingStrategy;
import org.datanucleus.store.ldap.fieldmanager.FetchFieldManager;
import org.datanucleus.util.ClassUtils;
import org.datanucleus.util.Localiser;
import org.datanucleus.util.NucleusLogger;

/**
 * Convenience methods for LDAP datastores.
 */
public class LDAPUtils
{
    /** Localiser for messages. */
    protected static final Localiser LOCALISER = Localiser.getInstance("org.datanucleus.store.ldap.Localisation", LDAPStoreManager.class
            .getClassLoader());

    public static final String[] NO_ATTRIBUTES = new String[0];

    /**
     * Gets the state manager for the given object.
     * @param object the object
     * @param om the object manager
     * @param persist true to persist the object if no state manager exists yet.
     * @return the state manager for object
     */
    public static ObjectProvider getStateManagerForObject(Object object, ExecutionContext om, boolean persist)
    {
        ObjectProvider smpc = om.findObjectProvider(object, persist);
        return smpc;
    }

    /**
     * Gets the effective class meta data for the given field meta data. This is either the class meta data of field
     * itself or, if the field is a collection type, the class meta data of the collection's elements.
     * @param mmd the field meta data
     * @param mmgr Metadata manager
     * @return the effective class meta data, null if none found
     */
    public static AbstractClassMetaData getEffectiveClassMetaData(AbstractMemberMetaData mmd, MetaDataManager mmgr)
    {
        ClassLoaderResolver clr = mmgr.getNucleusContext().getClassLoaderResolver(null);

        String[] fieldTypes = mmd.getFieldTypes();
        if(fieldTypes != null && fieldTypes.length > 0)
        {
            AbstractClassMetaData fieldTypeCmd = mmgr.getMetaDataForClass(fieldTypes[0], clr);
            if (fieldTypeCmd != null)
            {
                return fieldTypeCmd;
            }
        }

        AbstractClassMetaData fieldTypeCmd = mmgr.getMetaDataForClass(mmd.getType(), clr);
        if (fieldTypeCmd != null)
        {
            return fieldTypeCmd;
        }

        AbstractClassMetaData collectionFieldTypeCmd =
            (mmd.getCollection() != null) ? mmd.getCollection().getElementClassMetaData(clr, mmgr) : null;
        if (collectionFieldTypeCmd != null)
        {
            return collectionFieldTypeCmd;
        }

        return null;
    }

    /*
     * Helpers for hierarchical mapping
     */

    /**
     * Gets the member metadata from the given class metadata which attribute name matches the given LDAP attribute
     * name.
     * @param cmd the class metadata
     * @param attributeName the LDAP attribute name
     * @return the member metadata which attribute name matches, null if not found
     */
    public static AbstractMemberMetaData getMemberMetadataForAttributeName(AbstractClassMetaData cmd, String attributeName)
    {
        if (cmd != null && attributeName != null)
        {
            List<AbstractMemberMetaData> allMemberMetaData = getAllMemberMetaData(cmd);
            for (AbstractMemberMetaData mmd : allMemberMetaData)
            {
                String attributeNameForField = getAttributeNameForField(mmd);
                if (attributeName.equals(attributeNameForField))
                {
                    return mmd;
                }
            }
        }
        return null;
    }

    public static List<AbstractMemberMetaData> getAllMemberMetaData(AbstractClassMetaData cmd)
    {
        int[] fieldNumbers = cmd.getAllMemberPositions();
        return getMemberMetaData(fieldNumbers, cmd);
    }

    public static List<AbstractMemberMetaData> getMemberMetaData(int[] fieldNumbers, AbstractClassMetaData cmd)
    {
        List<AbstractMemberMetaData> mmds = new ArrayList<AbstractMemberMetaData>();
        if (fieldNumbers != null && fieldNumbers.length > 0)
        {
            for (int fieldNumber : fieldNumbers)
            {
                AbstractMemberMetaData mmd = cmd.getMetaDataForManagedMemberAtAbsolutePosition(fieldNumber);
                mmds.add(mmd);
            }
        }
        return mmds;
    }

    public static boolean isEmbeddedField(AbstractMemberMetaData mmd)
    {
        return mmd.getEmbeddedMetaData() != null || (mmd.getElementMetaData() != null && mmd.getElementMetaData().getEmbeddedMetaData() != null);
    }
    
    /**
     * Convenience method to check if the given object is mapped hierarchical.
     * @param sm StateManager
     * @return true if is hierarchical mapped
     */
    public static boolean isHierarchicalMappedAtChild(ObjectProvider sm)
    {
        LocationInfo locationInfo = getLocationInfo(sm.getClassMetaData());
        return locationInfo.parentFieldName != null;
    }

    /**
     * Gets the distinguished name of the parent object
     * @param dn The distinguished name
     * @param suffix the suffix defined in "dn" extension
     * @return the distinguished name of the parent object
     * @throws InvalidNameException
     */
    public static LdapName getParentDistingueshedName(LdapName dn, LdapName suffix)
    {
        LdapName parent = (LdapName) dn.getPrefix(dn.size() - suffix.size() - 1);
        return parent;
    }

    /**
     * Gets the relative distinguished name for object.
     * @param sm the state manager
     * @return the RDN for object
     */
    private static Rdn getRdnForObject(StoreManager storeMgr, ObjectProvider sm) throws InvalidNameException
    {
        AbstractClassMetaData cmd = sm.getClassMetaData();
        int fieldNumber = cmd.getPKMemberPositions()[0];
        Object value = sm.provideField(fieldNumber);
        AbstractMemberMetaData mmd = cmd.getMetaDataForManagedMemberAtAbsolutePosition(fieldNumber);
        Attributes rdnAttributes = new BasicAttributes();
        AbstractMappingStrategy ms = AbstractMappingStrategy.findMappingStrategy(storeMgr, sm, mmd, rdnAttributes);
        ms.insert(value);
        Rdn rdn = new Rdn(rdnAttributes);
        return rdn;
    }

    /**
     * Convenience method to return the distinguished name for the object being managed. Uses the extension "dn" if
     * specified, else the "table" if specified as parent container. Uses the PK name and value as RDN.
     * @param sm the StateManager
     * @return Distinguished name
     */
    public static LdapName getDistinguishedNameForObject(StoreManager storeMgr, ObjectProvider sm)
    {
        return getDistinguishedNameForObject(storeMgr, sm, false);
    }

    /**
     * Convenience method to return the distinguished name for the object being managed. Uses the extension "dn" if
     * specified, else the "table" if specified as parent container. Uses the PK name and value as RDN.
     * @param sm the StateManager
     * @param forceFetchHierarchicalMappedDn true to fetch the name from directory server
     * @return Distinguished name
     */
    public static LdapName getDistinguishedNameForObject(StoreManager storeMgr, ObjectProvider sm, boolean forceFetchHierarchicalMappedDn)
    {
        return getDistinguishedNameForObject(storeMgr, sm, null, forceFetchHierarchicalMappedDn);
    }

    private static LdapName getDistinguishedNameForObject(StoreManager storeMgr, ObjectProvider sm, Set<ObjectProvider> handledStateManagers,
            boolean forceFetchHierarchicalMappedDn)
    {
        if (handledStateManagers == null)
        {
            handledStateManagers = new HashSet<ObjectProvider>();
        }
        if (handledStateManagers.contains(sm))
        {
            throw new NucleusDataStoreException("Recursive loop detected while creating distinguished name for " + sm);
        }
        handledStateManagers.add(sm);

        AbstractClassMetaData cmd = sm.getClassMetaData();
        LocationInfo locationInfo = getLocationInfo(cmd);

        LdapName dn;
        try
        {
            SearchControls searchControls = getSearchControls(cmd);
            if (sm.getEmbeddedOwners() != null && sm.getEmbeddedOwners().length > 0)
            {
                ObjectProvider owner = sm.getEmbeddedOwners()[0];
                dn = getDistinguishedNameForObject(storeMgr, owner, handledStateManagers, forceFetchHierarchicalMappedDn);
                dn.add(getRdnForObject(storeMgr, sm));
            }
            else if (searchControls.getSearchScope() == SearchControls.OBJECT_SCOPE)
            {
                dn = locationInfo.dn;
            }
            else if (isHierarchicalMappedAtChild(sm))
            {
                Rdn rdn = getRdnForObject(storeMgr, sm);
                LdapName parentDn;
                AbstractMemberMetaData parentFieldMmd = cmd.getMetaDataForMember(locationInfo.parentFieldName);
                if (forceFetchHierarchicalMappedDn)
                {
                    // search for DN, use parent-DN as base
                    AbstractClassMetaData parentFieldTypeCmd = sm.getExecutionContext().getMetaDataManager().getMetaDataForClass(parentFieldMmd.getType(),
                        sm.getExecutionContext().getClassLoaderResolver());
                    LdapName base = getSearchBase(parentFieldTypeCmd, sm.getExecutionContext().getMetaDataManager());
                    String ocFilter = getSearchFilter(cmd);
                    String rdnFilter = "(" + rdn.getType() + "=" + rdn.getValue() + ")";
                    String filter = ocFilter != null ? "(&" + ocFilter + rdnFilter + ")" : rdnFilter;
                    ExecutionContext om = sm.getExecutionContext();
                    ManagedConnection mconn = storeMgr.getConnection(om);
                    try
                    {
                        DirContext ctx = (DirContext) mconn.getConnection();
                        if (NucleusLogger.DATASTORE_RETRIEVE.isDebugEnabled())
                        {
                            NucleusLogger.DATASTORE_RETRIEVE.debug(LOCALISER.msg("LDAP.JNDI.search", base, filter, searchControls
                                    .getSearchScope()));
                        }
                        NamingEnumeration<SearchResult> enumeration = ctx.search(base, filter, searchControls);
                        if (enumeration.hasMoreElements())
                        {
                            SearchResult sr = enumeration.nextElement();
                            String srName = sr.getNameInNamespace();
                            LdapName srDn = new LdapName(srName);
                            parentDn = getParentDistingueshedName(srDn, locationInfo.suffix);
                            enumeration.close();
                        }
                        else
                        {
                            if (locationInfo.dn != null)
                            {
                                parentDn = locationInfo.dn;
                            }
                            else
                            {
                                throw new NucleusObjectNotFoundException("No distinguished name found for object " + sm.toString());
                            }
                        }
                    }
                    catch (NamingException e)
                    {
                        throw new NucleusDataStoreException(e.getMessage(), e);
                    }
                    finally
                    {
                        mconn.release();
                    }
                }
                else
                {
                    Object parentFieldValue = sm.provideField(parentFieldMmd.getAbsoluteFieldNumber());
                    if (parentFieldValue != null)
                    {
                        // compose DN using parent DN
                        ExecutionContext om = sm.getExecutionContext();
                        boolean detached = om.getApiAdapter().isDetached(parentFieldValue);
                        ObjectProvider parentSm = getStateManagerForObject(parentFieldValue, om, detached);
                        if (parentSm == null)
                        {
                            throw new NucleusObjectNotFoundException("No state manager found for object " + parentFieldValue);
                        }
                        parentDn = getDistinguishedNameForObject(storeMgr, parentSm, handledStateManagers, forceFetchHierarchicalMappedDn);
                    }
                    else if (locationInfo.dn != null)
                    {
                        parentDn = locationInfo.dn;
                    }
                    else
                    {
                        // no way to get the DN from the object, fetch it from directory
                        handledStateManagers.remove(sm);
                        LdapName smDn = getDistinguishedNameForObject(storeMgr, sm, handledStateManagers, true);
                        parentDn = getParentDistingueshedName(smDn, locationInfo.suffix);
                    }
                }

                dn = composeDistinguishedName(parentDn, rdn, locationInfo.suffix);
            }
            else
            {
                dn = new LdapName(locationInfo.dn.toString());
                dn.add(getRdnForObject(storeMgr, sm));
            }
        }
        catch (InvalidNameException e)
        {
            throw new NucleusDataStoreException(e.getMessage(), e);
        }

        return dn;
    }

    public static LdapName getSearchBase(AbstractClassMetaData cmd, MetaDataManager mmgr)
    {
        LdapName dn;
        try
        {
            LocationInfo locationInfo = getLocationInfo(cmd);
            if (locationInfo.parentFieldName != null && locationInfo.dn == null)
            {
                AbstractMemberMetaData parentMmd = cmd.getMetaDataForMember(locationInfo.parentFieldName);
                ClassLoaderResolver clr = mmgr.getNucleusContext().getClassLoaderResolver(null);
                AbstractClassMetaData parentFieldCmd = mmgr.getMetaDataForClass(parentMmd.getType(), clr);
                LdapName parentDn = getSearchBase(parentFieldCmd, mmgr);

                dn = composeDistinguishedName(parentDn, null, locationInfo.suffix);
            }
            else
            {
                dn = locationInfo.dn;
            }
        }
        catch (InvalidNameException e)
        {
            throw new NucleusDataStoreException(e.getMessage(), e);
        }
        return dn;
    }

    /**
     * Gets the search controls. Contains scope SUBTREE_SCOPE in case hierarchical mapping is used, scope ONELEVEL_SCOPE
     * otherwise.
     * @param cmd the class meta data
     * @return the search controls
     */
    public static SearchControls getSearchControls(AbstractClassMetaData cmd)
    {
        SearchControls searchControls = new SearchControls();

        LocationInfo locationInfo = getLocationInfo(cmd);
        if (locationInfo.parentFieldName != null)
        {
            searchControls.setSearchScope(SearchControls.SUBTREE_SCOPE);
        }

        // overwrite scope if present in an LDAP URL
        if (locationInfo.scope != -1)
        {
            searchControls.setSearchScope(locationInfo.scope);
        }

        return searchControls;
    }

    /**
     * Convenience method to get the LDAP filter for the specified class.
     * @param cmd the class meta data
     * @return the object classes filter the specified class, null if no object class
     */
    public static String getSearchFilter(AbstractClassMetaData cmd)
    {
        // check for a filter in the LDAP URL
        LocationInfo locationInfo = getLocationInfo(cmd);
        String urlFilter = locationInfo.filter;

        Set<String> objectClasses = getObjectClassesForClass(cmd);
        if (objectClasses.isEmpty())
        {
            return null;
        }

        StringBuffer filter = new StringBuffer();
        if (objectClasses.size() > 1 || urlFilter != null)
        {
            filter.append("(&");
            for (String oc : objectClasses)
            {
                filter.append("(objectClass=");
                filter.append(oc);
                filter.append(")");
            }
            if (urlFilter != null)
            {
                filter.append(urlFilter);
            }
            filter.append(")");
        }
        else
        {
            filter.append("(objectClass=");
            filter.append(objectClasses.iterator().next());
            filter.append(")");
        }
        return filter.toString();
    }

    /**
     * Convenience method to return the distinguished name pattern for the metadata being managed. Uses the extension
     * "dn" if specified, else the "table" if specified.
     * @param cmd the class meta data
     * @return the DN as String
     */

    public static LocationInfo getLocationInfo(AbstractClassMetaData cmd)
    {
        String raw = null;
        if (cmd != null && cmd.hasExtension("dn"))
        {
            raw = cmd.getValueForExtension("dn");
        }
        else if (cmd != null && cmd.getTable() != null)
        {
            raw = cmd.getTable();
        }

        LocationInfo li = new LocationInfo();

        // parse the string
        // TODO: use org.apache.directory.shared.ldap.util.LdapURL class?
        if (raw != null)
        {
            String dnOrParentField = null;
            if (raw.startsWith("ldap:///"))
            {
                // format: ldap:///dn?attributes?scope?filter?extensions
                raw = raw.substring("ldap:///".length());
                String[] split = raw.split("\\?", 5);
                // dn
                if (split.length > 0)
                {
                    dnOrParentField = split[0];
                }
                // attributes
                if (split.length > 1)
                {
                    // ignore
                }
                // scope
                if (split.length > 2)
                {
                    String scope = split[2];
                    if (scope.length() > 0)
                    {
                        if ("base".equals(scope))
                        {
                            li.scope = SearchControls.OBJECT_SCOPE;
                        }
                        else if ("one".equals(scope))
                        {
                            li.scope = SearchControls.ONELEVEL_SCOPE;
                        }
                        else if ("sub".equals(scope))
                        {
                            li.scope = SearchControls.SUBTREE_SCOPE;
                        }
                        else
                        {
                            throw new NucleusDataStoreException("Invalid scope in LDAP URL: " + scope);
                        }
                    }
                }
                // filter
                if (split.length > 3)
                {
                    String filter = split[3];
                    if (filter.length() > 0)
                    {
                        li.filter = filter;
                    }
                }
            }
            else
            {
                dnOrParentField = raw;
            }

            if (dnOrParentField != null)
            {
                int left = dnOrParentField.indexOf('{');
                int right = dnOrParentField.indexOf('}');
                if (left > -1 && right > left)
                {
                    li.parentFieldName = dnOrParentField.substring(left + 1, right);

                    try
                    {
                        String suffix = dnOrParentField.substring(0, left);
                        LdapName suffixDn = new LdapName(suffix);
                        if (suffixDn.size() > 0 && suffixDn.getRdn(0).size() == 0)
                        {
                            suffixDn.remove(0);
                        }
                        li.suffix = suffixDn;
                    }
                    catch (InvalidNameException e)
                    {
                        throw new NucleusDataStoreException("Invalid LDAP DN: " + dnOrParentField);
                    }

                    if (dnOrParentField.length() > right + 2 && dnOrParentField.charAt(right + 1) == '|')
                    {
                        try
                        {
                            li.dn = new LdapName(dnOrParentField.substring(right + 2, dnOrParentField.length()));
                        }
                        catch (InvalidNameException e)
                        {
                            throw new NucleusDataStoreException("Invalid LDAP DN: " + dnOrParentField);
                        }
                    }
                }
                else
                {
                    try
                    {
                        li.dn = new LdapName(dnOrParentField);
                    }
                    catch (InvalidNameException e)
                    {
                        throw new NucleusDataStoreException("Invalid LDAP DN: " + dnOrParentField);
                    }
                }
            }

        }

        return li;
    }

    /**
     * Information where an object is located in the LDAP tree.
     */
    public static class LocationInfo
    {
        /** The LDAP distinguished name, null if none */
        public LdapName dn = null;

        /** The parent field name, null if none */
        public String parentFieldName = null;

        /** The suffix, only if parent field name is set, null if none */
        public LdapName suffix = null;

        /** The filter if the LDAP URL, null if none */
        public String filter = null;

        /** The search scope, -1 if none */
        public int scope = -1;
    }

    /**
     * Convenience method to return value defined in extension "empty-value" of the given field meta data, null if not
     * specified. Used as workaround for attributes member (object class groupOfNames) and uniqueMemeber (object
     * classgroupOfUniqueNames) which must contain at least one value
     * @param mmd the member meta data
     * @return the value
     */
    public static String getEmptyValue(AbstractMemberMetaData mmd)
    {
        return mmd.getValueForExtension("empty-value");
    }

    /*
     * Helper for MetaData <-> LDAP translation
     */

    /**
     * Gets the attribute value of an specific attribute name from the state manager.
     * @param sm the state manager
     * @param attributeName the attribute name
     * @return the attribute value
     */
    public static Object getAttributeValue(StoreManager storeMgr, ObjectProvider sm, String attributeName)
    {
        try
        {
            // get field value of the PC
            AbstractMemberMetaData pcMmd = LDAPUtils.getMemberMetadataForAttributeName(sm.getClassMetaData(), attributeName);
            Object pcFieldValue = sm.provideField(pcMmd.getAbsoluteFieldNumber());

            // get LDAP value
            Attributes pcAttributes = new BasicAttributes();
            AbstractMappingStrategy ms = AbstractMappingStrategy.findMappingStrategy(storeMgr, sm, pcMmd, pcAttributes);
            ms.insert(pcFieldValue);
            Attribute pcAttribute = pcAttributes.get(attributeName);
            Object pcAttributeValue = pcAttribute.get();

            return pcAttributeValue;
        }
        catch (NamingException e)
        {
            throw new NucleusDataStoreException(e.getMessage(), e);
        }
    }

    /**
     * Gets the attribute value of an specific attribute name from LDAP.
     * @param sm the state manager
     * @param attributeName the attribute name
     * @return the attribute value
     */
    public static Collection<Object> getAttributeValuesFromLDAP(StoreManager storeMgr, ObjectProvider sm, String attributeName)
    {
        ManagedConnection mconn = storeMgr.getConnection(sm.getExecutionContext());
        try
        {
            LdapName dn = getDistinguishedNameForObject(storeMgr, sm, true);
            DirContext ctx = (DirContext) mconn.getConnection();
            if (NucleusLogger.DATASTORE_RETRIEVE.isDebugEnabled())
            {
                NucleusLogger.DATASTORE_RETRIEVE.debug(LOCALISER.msg("LDAP.JNDI.getAttributes", dn, attributeName, ""));
            }
            Attributes attributes = ctx.getAttributes(dn, new String[]{attributeName});
            Attribute attribute = attributes.get(attributeName);
            Collection<Object> pcAttributeValues = new ArrayList<Object>();
            if (attribute != null)
            {
                NamingEnumeration<?> all = attribute.getAll();
                while (all.hasMoreElements())
                {
                    pcAttributeValues.add(all.nextElement());
                }
            }
            return pcAttributeValues;
        }
        catch (NamingException e)
        {
            throw new NucleusDataStoreException(e.getMessage(), e);
        }
        finally
        {
            mconn.release();
        }
    }

    /**
     * Accessor for the (LDAP) attribute name for the specified field. Uses the extension "dn" if specified
     * (deprecated), else the extension "attribute" if specified, else the "column" if specified, else uses the field
     * name.
     * @param mmd MetaData for the field
     * @return The attribute
     */
    public static String getAttributeNameForField(AbstractMemberMetaData mmd)
    {
        String name = mmd.getName();
        if (mmd.hasExtension("dn"))
        {
            name = mmd.getValueForExtension("dn");
        }
        else if (mmd.hasExtension("attribute"))
        {
            name = mmd.getValueForExtension("attribute");
        }
        else if (mmd.getColumnMetaData() != null && mmd.getColumnMetaData().length > 0)
        {
            name = mmd.getColumnMetaData()[0].getName();
        }
        return name;
    }

    /**
     * Accessor for the (LDAP) object classes for the specified class. Uses the extension "objectClass" if specified,
     * else the "schema" if specified. name.
     * @param sm StateManager
     * @return The object classes
     */
    public static Set<String> getObjectClassesForClass(AbstractClassMetaData cmd)
    {
        Set<String> ocs = new LinkedHashSet<String>();
        if (cmd.hasExtension("objectClass"))
        {
            ocs.addAll(Arrays.asList(cmd.getValuesForExtension("objectClass")));
        }
        else if (cmd.getSchema() != null)
        {
            ocs.addAll(Arrays.asList(MetaDataUtils.getInstance().getValuesForCommaSeparatedAttribute(cmd.getSchema())));
        }
        return ocs;
    }

    /**
     * Composes the distinguished name.
     * @param parentDn the parent dn
     * @param rdn the rdn
     * @param suffix the suffix
     * @return the ldap name
     * @throws InvalidNameException
     */
    public static LdapName composeDistinguishedName(LdapName parentDn, Rdn rdn, LdapName suffix) throws InvalidNameException
    {
        LdapName dn = new LdapName(parentDn.getRdns());
        dn.addAll(suffix.getRdns());
        if (rdn != null)
        {
            dn.add(rdn);
        }
        return dn;
    }

    public static Object getObjectByDN(StoreManager storeMgr, ExecutionContext om, Class type, String dnAsString)
    {
        AbstractClassMetaData cmd = om.getMetaDataManager().getMetaDataForClass(type, om.getClassLoaderResolver());
        LdapName base = getSearchBase(cmd, om.getMetaDataManager());

        try
        {
            LdapName dn = new LdapName(dnAsString);
            Rdn rdn = dn.getRdn(dn.size() - 1);
            String filter = "(" + rdn.getType() + "=" + rdn.getValue() + ")";
            List<Object> objects = LDAPUtils.getObjectsOfCandidateType(storeMgr, om, cmd, base, filter, true, false);
            if (objects.size() == 1)
            {
                return objects.get(0);
            }
            else if (objects.size() == 0)
            {
                // TODO: localise
                throw new NucleusObjectNotFoundException("No object found with type " + type + " and filter " + filter);
            }
            else
            {
                // TODO: localise
                throw new NucleusDataStoreException("Unambiguous match with type " + type + " and filter " + filter);
            }
        }
        catch (NamingException e)
        {
            throw new NucleusDataStoreException(e.getMessage(), e);
        }
    }

    /**
     * Gets an object by type and attribute.
     * @param om object manager
     * @param type the object type
     * @param attributeName the attribute name
     * @param attributeValue the attribute value
     * @param mmgr the meta data manager
     * @return The object
     */
    public static Object getObjectByAttribute(StoreManager storeMgr, ExecutionContext om, Class type, String attributeName, String attributeValue,
            MetaDataManager mmgr)
    {
        String attributeFilter = "(" + attributeName + "=" + attributeValue + ")";
        AbstractClassMetaData cmd = om.getMetaDataManager().getMetaDataForClass(type, om.getClassLoaderResolver());
        LdapName base = getSearchBase(cmd, mmgr);
        List<Object> objects = LDAPUtils.getObjectsOfCandidateType(storeMgr, om, cmd, base, attributeFilter, true, false);
        if (objects.size() == 1)
        {
            return objects.get(0);
        }
        else if (objects.size() == 0)
        {
            // TODO: localise
            throw new NucleusObjectNotFoundException("No object found with type " + type + " and filter " + attributeFilter);
        }
        else
        {
            // TODO: localise
            throw new NucleusDataStoreException("Unambiguous match with type " + type + " and filter " + attributeFilter);
        }
    }

    public static void markForPersisting(Object pc, ExecutionContext om)
    {
        LDAPTransactionEventListener listener = getTransactionEventListener(om);
        listener.addObjectToPersist(pc);
    }

    public static void markForRename(StoreManager storeMgr, Object pc, ExecutionContext om, LdapName oldDn, LdapName newDn)
    {
        ManagedConnection mconn = storeMgr.getConnection(om);
        try
        {
            DirContext ctx = (DirContext) mconn.getConnection();
            if (NucleusLogger.DATASTORE_PERSIST.isDebugEnabled())
            {
                NucleusLogger.DATASTORE_PERSIST.debug(LOCALISER.msg("LDAP.JNDI.rename", oldDn, newDn));
            }
            ctx.rename(oldDn, newDn);
        }
        catch (NamingException e)
        {
            throw new NucleusDataStoreException(e.getMessage(), e);
        }
        finally
        {
            mconn.release();
        }
    }

    public static void markForDeletion(Object pc, ExecutionContext om)
    {
        LDAPTransactionEventListener listener = getTransactionEventListener(om);
        listener.addObjectToDelete(pc);
    }

    public static void unmarkForDeletion(Object pc, ExecutionContext om)
    {
        LDAPTransactionEventListener listener = getTransactionEventListener(om);
        listener.removeObjectToDelete(pc);
    }

    private static LDAPTransactionEventListener getTransactionEventListener(ExecutionContext om)
    {
        Transaction transaction = om.getTransaction();
        LDAPTransactionEventListener listener = (LDAPTransactionEventListener) transaction.getOptions().get("LDAPTransactionEventListener");
        if (listener == null)
        {
            listener = new LDAPTransactionEventListener(om);
            transaction.getOptions().put("LDAPTransactionEventListener", listener);
            transaction.addTransactionEventListener(listener);
        }
        return listener;
    }

    /*
     * Search methods
     */

    /**
     * Convenience method to get all objects of the candidate type (and optional subclasses).
     * @param om The object manager
     * @param compilation The query
     * @param parameters The input parameters
     * @param candidateClass The class of the candidates
     * @param subclasses Include subclasses?
     * @param ignoreCache Whether to ignore the cache
     * @param inMemory Whether to filter in memory or to use native LDAP filters
     * @return List of objects of the candidate type (or subclass)
     */
    public static List<Object> getObjectsOfCandidateType(StoreManager storeMgr, ExecutionContext om, QueryCompilation compilation, Map parameters,
            Class candidateClass, boolean subclasses, boolean ignoreCache, boolean inMemory)
    {
        ClassLoaderResolver clr = om.getClassLoaderResolver();
        AbstractClassMetaData cmd = om.getMetaDataManager().getMetaDataForClass(candidateClass, clr);
        List<Object> results = getObjectsOfCandidateType(storeMgr, om, compilation, parameters, cmd, ignoreCache, inMemory);
        if (subclasses)
        {
            // Add on any subclass objects
            String[] subclassNames = om.getMetaDataManager().getSubclassesForClass(candidateClass.getName(), true);
            if (subclassNames != null)
            {
                for (String subclassName : subclassNames)
                {
                    AbstractClassMetaData subCmd = om.getMetaDataManager().getMetaDataForClass(subclassName, clr);
                    results.addAll(getObjectsOfCandidateType(storeMgr, om, compilation, parameters, subCmd, ignoreCache, inMemory));
                }
            }
        }
        return results;
    }

    private static List<Object> getObjectsOfCandidateType(StoreManager storeMgr, ExecutionContext om, QueryCompilation compilation, Map parameters,
            final AbstractClassMetaData acmd, boolean ignoreCache, boolean inMemory)
    {
        List<Object> results = new ArrayList<Object>();

        String classFilter = getSearchFilter(acmd);
        if (classFilter != null) // this is a sanity check, if the class does not have the objectClass
        {
            if (!inMemory)
            {
                try
                {
                    // this class cannot be linked at compile time since its dependencies are considered as optional
                    String className = "org.datanucleus.store.ldap.query.QueryToLDAPFilterMapper";
                    Class cls = om.getClassLoaderResolver().classForName(className);
                    Method method = ClassUtils.getMethodForClass(cls, "compile", null);
                    Constructor constr = ClassUtils.getConstructorWithArguments(cls, new Class[]{QueryCompilation.class, Map.class,
                            AbstractClassMetaData.class});
                    String filter = (String) method
                            .invoke(constr.newInstance(new Object[]{compilation, parameters, acmd}), (Object[]) null);

                    results = getObjectsOfCandidateType(storeMgr, om, acmd, null, filter, ignoreCache);
                }
                catch (Throwable e)
                {
                    NucleusLogger.QUERY.warn(LOCALISER.msg("LDAP.Query.NativeQueryFailed"));
                    // on error switch back to in-memory handling
                    inMemory = true;
                }
            }
            if (inMemory)
            {
                results = getObjectsOfCandidateType(storeMgr, om, acmd, null, classFilter, ignoreCache);
            }
        }
        return results;
    }

    /**
     * @param om The object manager
     * @param candidateCmd The class meta data of the candidate
     * @param base the search base, or null to use the base of candidateCmd
     * @param additionalFilter the additional filter, or null to only use the filter of candidateCmd
     * @param subclasses true to include subclasses
     * @param ignoreCache whether to ignore the cache
     * @return List of objects
     */
    public static List<Object> getObjectsOfCandidateType(StoreManager storeMgr, ExecutionContext om, final AbstractClassMetaData candidateCmd, LdapName base,
            String additionalFilter, boolean subclasses, boolean ignoreCache)
    {
        ClassLoaderResolver clr = om.getClassLoaderResolver();
        List<Object> results = getObjectsOfCandidateType(storeMgr, om, candidateCmd, base, additionalFilter, ignoreCache);
        if (subclasses)
        {
            // Add on any subclass objects
            String[] subclassNames = om.getMetaDataManager().getSubclassesForClass(candidateCmd.getFullClassName(), true);
            if (subclassNames != null)
            {
                for (String subclassName : subclassNames)
                {
                    AbstractClassMetaData subCmd = om.getMetaDataManager().getMetaDataForClass(subclassName, clr);
                    results.addAll(getObjectsOfCandidateType(storeMgr, om, subCmd, base, additionalFilter, ignoreCache));
                }
            }
        }
        return results;
    }

    private static List<Object> getObjectsOfCandidateType(final StoreManager storeMgr, final ExecutionContext om, final AbstractClassMetaData cmd, 
        LdapName base, String additionalFilter, boolean ignoreCache)
    {
        List<Object> results = new ArrayList<Object>();

        final ClassLoaderResolver clr = om.getClassLoaderResolver();
        Map<LdapName, Attributes> entries = getEntries(storeMgr, om, cmd, base, additionalFilter, false, ignoreCache);
        for (final Attributes attrs : entries.values())
        {
            results.add(om.findObjectUsingAID(new Type(clr.classForName(cmd.getFullClassName())), new FieldValues()
            {
                // StateManager calls the fetchFields method
                public void fetchFields(ObjectProvider sm)
                {
                    sm.replaceFields(cmd.getPKMemberPositions(), new FetchFieldManager(storeMgr, sm, attrs));
                    
                    List<AbstractMemberMetaData> allMemberMetaData = getAllMemberMetaData(cmd);
                    List<AbstractMemberMetaData> basicMemberMetaData = new ArrayList<AbstractMemberMetaData>();
                    for (AbstractMemberMetaData mmd : allMemberMetaData)
                    {
                        mmd.getAbsoluteFieldNumber();
                        if (mmd.getRelationType(clr) == Relation.NONE &&
                            !mmd.isPersistentInterface(clr, om.getMetaDataManager()) && 
                            !Collection.class.isAssignableFrom(mmd.getType()) && !Map.class.isAssignableFrom(mmd.getType()) && !mmd.getType().isArray())
                        {
                            basicMemberMetaData.add(mmd);
                        }
                    }
                    int[] basicMemberPosition = new int[basicMemberMetaData.size()];
                    for (int i = 0; i < basicMemberMetaData.size(); i++)
                    {
                        basicMemberPosition[i] = basicMemberMetaData.get(i).getAbsoluteFieldNumber();
                    }
                    sm.replaceFields(basicMemberPosition, new FetchFieldManager(storeMgr, sm, attrs));
                }

                public void fetchNonLoadedFields(ObjectProvider sm)
                {
                    sm.replaceNonLoadedFields(cmd.getAllMemberPositions(), new FetchFieldManager(storeMgr, sm, attrs));
                }

                public FetchPlan getFetchPlanForLoading()
                {
                    return null;
                }
            }, ignoreCache, true));
        }

        return results;
    }

    public static Map<LdapName, Attributes> getEntries(StoreManager storeMgr, ExecutionContext om, final AbstractClassMetaData candidateCmd, LdapName base,
            String additionalFilter, boolean subclasses, boolean ignoreCache)
    {
        ManagedConnection mconn = storeMgr.getConnection(om);
        try
        {
            DirContext ctx = (DirContext) mconn.getConnection();
            ClassLoaderResolver clr = om.getClassLoaderResolver();
            Map<LdapName, Attributes> results = getEntries(om, candidateCmd, ctx, base, additionalFilter, ignoreCache);
            if (subclasses)
            {
                // Add on any subclass objects
                String[] subclassNames = om.getMetaDataManager().getSubclassesForClass(candidateCmd.getFullClassName(), true);
                if (subclassNames != null)
                {
                    for (String subclassName : subclassNames)
                    {
                        AbstractClassMetaData cmd = om.getMetaDataManager().getMetaDataForClass(subclassName, clr);
                        results.putAll(getEntries(om, cmd, ctx, base, additionalFilter, ignoreCache));
                    }
                }
            }
            return results;
        }
        finally
        {
            mconn.release();
        }
    }

    private static Map<LdapName, Attributes> getEntries(ExecutionContext om, final AbstractClassMetaData acmd, DirContext ctx, LdapName base,
            String additionalFilter, boolean ignoreCache)
    {
        Map<LdapName, Attributes> results = new LinkedHashMap<LdapName, Attributes>();

        // no results for interface or abstract
        Class c = om.getClassLoaderResolver().classForName(acmd.getFullClassName());
        if (c.isInterface() || Modifier.isAbstract(c.getModifiers()))
        {
            return results;
        }

        try
        {
            if (base == null)
            {
                base = getSearchBase(acmd, om.getMetaDataManager());
            }
            String filter = getSearchFilter(acmd);
            if (additionalFilter != null)
            {
                filter = "(&" + filter + additionalFilter + ")";
            }
            SearchControls searchControls = getSearchControls(acmd);

            if (NucleusLogger.DATASTORE_RETRIEVE.isDebugEnabled())
            {
                NucleusLogger.DATASTORE_RETRIEVE.debug(LOCALISER.msg("LDAP.JNDI.search", base, filter, searchControls.getSearchScope()));
            }

            NamingEnumeration<SearchResult> enumeration = ctx.search(base.toString(), filter, searchControls);;
            while (enumeration.hasMoreElements())
            {
                final SearchResult sr = enumeration.nextElement();
                final Attributes attrs = sr.getAttributes();
                String name = sr.getNameInNamespace();
                LdapName dn = new LdapName(name);

                if (searchControls.getSearchScope() == SearchControls.SUBTREE_SCOPE)
                {
                    if (dn.equals(base))
                    {
                        // skip this entry, it is the search base
                        continue;
                    }
                }

                results.put(dn, attrs);
            }
        }
        catch (NameNotFoundException nnfe)
        {
            // ignore, occurs when trying to search for an non-existing object
        }
        catch (NamingException ne)
        {
            throw new NucleusDataStoreException(ne.getMessage(), ne);
        }

        return results;
    }

    public static void insert(StoreManager storeMgr, LdapName dn, Attributes attributes, ExecutionContext om)
    {
        if (NucleusLogger.DATASTORE_PERSIST.isDebugEnabled())
        {
            NucleusLogger.DATASTORE_PERSIST.debug(LOCALISER.msg("LDAP.JNDI.createSubcontext", dn, attributes));
        }
        ManagedConnection mconn = storeMgr.getConnection(om);
        try
        {
            DirContext ctx = (DirContext) mconn.getConnection();
            ctx.bind(dn, null, attributes);
        }
        catch (NamingException ne)
        {
            throw new NucleusDataStoreException(ne.getMessage(), ne);
        }
        finally
        {
            mconn.release();
        }
    }

    public static void update(StoreManager storeMgr, LdapName dn, Attributes attributes, ExecutionContext om)
    {
        if (NucleusLogger.DATASTORE_PERSIST.isDebugEnabled())
        {
            NucleusLogger.DATASTORE_PERSIST.debug(LOCALISER.msg("LDAP.JNDI.modifyAttributes", dn, "REPLACE", attributes));
        }
        ManagedConnection mconn = storeMgr.getConnection(om);
        try
        {
            DirContext ctx = (DirContext) mconn.getConnection();
            ctx.modifyAttributes(dn, DirContext.REPLACE_ATTRIBUTE, attributes);
        }
        catch (NamingException ne)
        {
            throw new NucleusDataStoreException(ne.getMessage(), ne);
        }
        finally
        {
            mconn.release();
        }
    }

    public static void deleteRecursive(StoreManager storeMgr, LdapName dn, ExecutionContext om)
    {
        ManagedConnection mconn = storeMgr.getConnection(om);
        try
        {
            DirContext ctx = (DirContext) mconn.getConnection();
            deleteRecursive(dn, ctx);
        }
        catch (NamingException ne)
        {
            throw new NucleusDataStoreException(ne.getMessage(), ne);
        }
        finally
        {
            mconn.release();
        }
    }

    public static void deleteRecursive(LdapName dn, DirContext ctx) throws NamingException
    {
        // search one-level and delete each
        // on exception: delete all children recursive
        NamingEnumeration<SearchResult> enumeration = ctx.search(dn, "(objectClass=*)", new SearchControls());
        while (enumeration.hasMoreElements())
        {
            SearchResult result = enumeration.nextElement();
            String resultName = result.getNameInNamespace();
            LdapName resultDn = new LdapName(resultName);
            try
            {
                if (NucleusLogger.DATASTORE_PERSIST.isDebugEnabled())
                {
                    NucleusLogger.DATASTORE_PERSIST.debug(LOCALISER.msg("LDAP.JNDI.destroySubcontext", resultDn));
                }
                ctx.unbind(resultDn);
            }
            catch (ContextNotEmptyException cnee)
            {
                deleteRecursive(resultDn, ctx);
            }
        }

        // now all child entries are deleted, delete parent again
        if (NucleusLogger.DATASTORE_PERSIST.isDebugEnabled())
        {
            NucleusLogger.DATASTORE_PERSIST.debug(LOCALISER.msg("LDAP.JNDI.destroySubcontext", dn));
        }
        ctx.unbind(dn);
    }

}